unit CobDecoderU;

interface
uses
  sysutils, 
  stdU, 
  StreamsU, 
  listsU, 
  Cob_Format;

type
  EScriptingError = class( Exception );

  TCobDecoder = class;
  TProcData = class;
  
  TCobByteCode = TArrayOfLongword;
  TOffsetArray = TArrayOfLongword;

  TCobProc = record
    Name : string;
    Offset : LongInt;
  end; {TCobProc}

  TProcData = class
  protected
    // property fields
    fProcName : string;
    fStream : IStream;
    fCobDecoder : TCobDecoder;
    // the list of addresses to decompile
    DecompileAddresses : TList;
    AddressesToVist : TStack;

    // Tests if a sting is an integer, if it Value contains the results
    // Does not throw an exception if it is invalid, returns false
    Function IsInteger( const s : string; var Value : Integer ) : boolean;
    // outputs a line of source code corosponding to the decompiled code.
    procedure OutputLine( const s : string;
                          const args : array of const);    
  public
    // the array of piece names
    PieceNames : TStringArray;
    // the array of Proc names
    ProcNames : TStringArray;
    // expression stack
    ExprStack : TStringlist;
    // the actual output
    Lines : TStringlist;
    // List of the parameters, initially every local variable is a parameter
    // till determined otherwise
    Params : TStringlist;
    // list of local variables for the script method
    Locals : TStringlist;
    // list of global variables for the script
    Globals : TStringlist;
    // Constructor Creates the script method decompiler object
    Constructor Create( aCobDecoder : TCobDecoder;
                        aStream : IStream;
                        aPieceNames, aProcNames : TStringArray;
                        const aProcName : String;
                        aGlobalVarsCount : Integer);
    Destructor Destroy; override;

    // gets an expression off the expression stack
    // If the expression getter, has a higher Priority than the expression,
    // the expression is wrapped in brackets to preserve the expression consistance
    // The result is the priority of the expression
    Function PopExpr( var Expr : string; ExprePriority : LongInt ) : Longint;
    // pushes an expression onto the expression stack.
    // The Expression priority is stored with the expression
    Procedure PushExpr( const Expr : string; ExprePriority : LongInt );
    // dumps the contents of the stack to the output
    // does not change the stack
    Procedure ExprStackDump;
    {
     Adds an address to decompile from to the list of address than require
     decompilialtion. Fails, if the address have been visited.
     This decompiler parses forward, when it encounters a conditional goto
     statment, it marks to target to be decompiled and continues to decompile.
     When it encounters the target it stops the decompiling loop. This allows
     various 'smart' formating to occur on the outputed source code
     If a return statment is ecountered before the target is found, decompilation stops.
     Decompilation then resumes at the next decompilation point(the target)
    }
    Procedure AddAddressToDecompileFrom( {StartOffset, End}Offset : Longint );
    // gets the next address to decompile
    Function GetNextAddressToDecompileFrom : LongInt;
    // Check to see if more addresses to decompile exist
    Function IsMoreAddressToDecompile : boolean;

    // Decompiles a method starting at a given offset
    Procedure Decompile( StartOffset : Longint );
    {
      Starts decompiling from an address.
      Decompilation stops when:
      - A Return statment is found.
      - Reaches
      - It runs out of stream to decompile.
      - An error occurs (an exception is raised)
    }
    Procedure DecodeFromAddress( StartOffset : Longint );
    // decomiles a single(most of the time) opcode. Does the gunt work
    Function DecodeOpcode( aOpCode : Longword ) : boolean;

    Property ProcName : string read fProcName write fProcName;
    Property Stream : IStream read fstream;
    Property CobDecoder : TCobDecoder read fCobDecoder;
  end; {TProcData}

  TCodeRefactoriser = class
  Public
    // the refactored source code
    Lines : TStringlist;
    Constructor Create(SourceLines, Globals ,Locals, Params : TStringlist); virtual; abstract;
  end; {TCodeRefactoriser}


  TCobDecoder = class
  Protected
    fScriptName : string;
    Stream : IStream;
    PieceNames : TStringArray;
    ProcNames : TStringArray;
    // raw bytecode for the script Procs
    Procs : array of TCobProc;
    GlobalVarsCount : Integer;
    // the raw byte code for the entire script
    ByteCode : TCobByteCode;
    Procedure ReadHeader;
    function ReadString( Offset : Longint ) : string;
    function ReadStringArray( Size, Offset : Longint ) : TStringArray;
    function ReadOffsetArray( Size, Offset : longint ) : TOffsetArray;
    Procedure ReadProcs( const CobHeader : TCobHeader );
    Procedure DecodeProcs;
    Procedure DecodeProc( stream : IStream;const CobProc : TCobProc );
  public
    // output
    OutputStream : IStringStream;
    Constructor Create( aStream : IStream;const aScriptName : string );
    Destructor Destroy; override;
    // the script name
    Property ScriptName : string read fScriptName;
  end; {TCobDecoder}

Procedure DecodeCob( stream : IStream; const ScriptName : string );


const
  Linear_Constant  = 163840;
  ANGULAR_CONSTANT = 182;

const
  // these constants control when an expression gets a pair of brackets wrapped around it.
  // Higher value, will cause the lower value expr to be wrapped in brackets
  ValueOp = 6;
  MulOp   = 5;
  NotOp   = 4;
  AddOP   = 3;
  RelOp   = 2;
  ResOp   = 1;


Type
  TUnitProperty =record
    Key : Longint;
    Name : string;
    Flags : set of (SetAble, ArgIsPiece, ArgIsUnitID, IsBool);
    // max of 5
    ArgCount : 0..4;
  end; {TUnitProperty}
  
// Indices for set/get value
const
  // set Or get
  Activation         =  1;
  // set Or get
  StandingMoveOrders =  2;
  // set Or get
  StandingFireOrders =  3;
  // get (0-100%)
  Health             =  4;
  // set Or get
  InBuildStance      =  5;
  // set Or get (used by misc. special case missions like transport ships)
  Busy               =  6;
  // get
  Piece_XZ           =  7;
  // get
  Piece_Y            =  8;
  // get
  Unit_XZ            =  9;
  // get
  Unit_Y             = 10;
  // get
  Unit_Height        = 11;
  // get atan of packed x,z coords
  XZ_aTan            = 12;
  // get hypot of packed x,z coords
  XZ_Hypot           = 13;
  // get ordinary two-parameter atan
  aTan               = 14;
  // get ordinary two-parameter hypot
  Hypot              = 15;
  // get
  Ground_Height      = 16;
  // get 0 = unit is built and ready, 1-100 = How much is left to build
  Build_Percent_Left = 17;
  // set Or get (change which plots we occupy when building opens and closes)
  Yard_Open          = 18;
  // set Or get (ask other units to clear the area)
  Bugger_Off         = 19;
  // set Or get
  Armored            = 20;

  // The prefix for a unit property
  UnitPropPrefix = 'UnitObj.';
  // various unit properties
  UnitProperties : array [ 1..20 ] of TUnitProperty = (
                 (key : Activation;            Name : 'Activation';
                  Flags : [ SetAble,IsBool ];  ArgCount : 0),
                 (key : StandingMoveOrders;    Name : 'StandingMoveOrders';
                  Flags : [ SetAble ];         ArgCount : 0),
                 (key : StandingFireOrders;    Name : 'StandingFireOrders';
                  Flags : [ SetAble ];         ArgCount : 0),
                 (key : Health;                Name : 'Health';
                  Flags : [ ];                  ArgCount : 0),
                 (key : InBuildStance;         Name : 'InBuildStance';
                  Flags : [ SetAble ];         ArgCount : 0),
                 (key : Busy;                  Name : 'Busy';
                  Flags : [ SetAble, IsBool ]; ArgCount : 0),
                 (key : Piece_XZ;              Name : 'Piece_XZ';
                  Flags : [ ArgIsPiece ];      ArgCount : 1),
                 (key : Piece_Y;               Name : 'Piece_Y';
                  Flags : [ ArgIsPiece ];      ArgCount : 1),
                 (key : Unit_XZ;               Name : 'Unit_XZ';
                  Flags : [ ArgIsUnitID ];     ArgCount : 1),
                 (key : Unit_Y;                Name : 'Unit_Y';
                  Flags : [ ArgIsUnitID ];     ArgCount : 1),
                 (key : Unit_Height;           Name : 'Unit_Height';
                  Flags : [ ArgIsUnitID ];     ArgCount : 1),
                 (key : XZ_aTan;               Name : 'XZ_aTan';
                  Flags : [ ];                  ArgCount : 1),
                 (key : XZ_Hypot;              Name : 'XZ_Hypot';
                  Flags : [ ];                  ArgCount : 1),
                 (key : aTan;                  Name : 'aTan';
                  Flags : [ ];                  ArgCount : 2),
                 (key : Hypot;                 Name : 'Hypot';
                  Flags : [ ];                  ArgCount : 2),
                 (key : Ground_Height;         Name : 'Ground_Height';
                  Flags : [ ];                  ArgCount : 1),
                 (key : Build_Percent_Left;    Name : 'Build_Percent_Left';
                  Flags : [ ];                  ArgCount : 0),
                 (key : Yard_Open;             Name : 'Yard_Open';
                  Flags : [ SetAble, IsBool ]; ArgCount : 0),
                 (key : Bugger_Off;            Name : 'Bugger_Off';
                  Flags : [ SetAble, IsBool ]; ArgCount : 0),
                 (key : Armored;               Name : 'Armored';
                  Flags : [ SetAble, IsBool ]; ArgCount : 0));

{
Special Effect Particles referenced in the scripting language
with the command "emit-sfx".  This file is included by any
scripts that use the command.
}
const
// Vector-based special effects

  SFXTYPE_VTOL              = 0;
  SFXTYPE_THRUST            = 1;
  SFXTYPE_WAKE1             = 2;
  SFXTYPE_WAKE2             = 3;
  SFXTYPE_REVERSEWAKE1      = 4;
  SFXTYPE_REVERSEWAKE2      = 5;

// Point-based (piece origin) special effects

  SFXTYPE_POINTBASED        = 256;

  SFXTYPE_WHITESMOKE        = (SFXTYPE_POINTBASED Or 1);
  SFXTYPE_BLACKSMOKE        = (SFXTYPE_POINTBASED Or 2);
  SFXTYPE_SUBBUBBLES        = (SFXTYPE_POINTBASED Or 3);

// SFX Occupy States
  SFXOCCUPY_AIR = 5;
{
Exploding pieces are activated in the scripting language with
the command "explode".  This file is included by any scripts
that use the command.
}

Const
  SHATTER        = 1;  // The piece will shatter instead of remaining whole
  EXPLODE_ON_HIT = 2;  // The piece will explode when it hits the ground
  FALL           = 4;  // The piece will fall due to gravity instead of just flying off
  SMOKE          = 8;  // A smoke trail will follow the piece through the air
  FIRE           = 16; // A fire trail will follow the piece through the air
  BITMAPONLY     = 32; // The piece will not fly off Or shatter Or anything.  Only a bitmap explosion will be rendered.


// Bitmap Explosion Types (these will be changed eventually)

  BITMAP1    = 256;
  BITMAP2    = 512;
  BITMAP3    = 1024;
  BITMAP4    = 2048;
  BITMAP5    = 4096;
  BITMAPNUKE = 8192;

  BITMAPMASK = 16128; // Mask of the possible bitmap bits

  NewLine = #10#13;
  
const
  // error msg
  UnitPropertyNotSetable = '%s%%s is not setable';
  InvaliePieceIndex = 'Invalid piece index to attach unit to';
  ExprStackUnderFlow = 'Expression stack underflow.'+NewLine+
                       'In Script "%s" Method "%s"'+NewLine+
                       'This is due to the script being corrupted Or there has been an internal error in the decompiler';
  UnknownSFXType = 'Unknown SFX type $%p';
Function BoolToStr( value : boolean ) : string;

implementation
uses
  LowLevelIO;

Function BoolToStr( value : boolean ) : string;
begin
if value then
  result := '1'
else
  result := '0';
end; {BoolToStr}

Procedure DecodeCob( stream : IStream; const ScriptName : string );
var
  CobDecoder : TCobDecoder;
begin
// create the decompiler object
CobDecoder := TCobDecoder.create( stream, ScriptName );
try
  // do stuff (todo)
{
  if CobDecoder.Output.DataString <> '' then
    begin
    writeln( CobDecoder.Output.DataString );
    readln;
    end;
}
finally // clean up
  CobDecoder.free;
end;
end; {DecodeCob}


// -----------------------------------------------------------------------------
//  TProcData
// -----------------------------------------------------------------------------

Constructor TProcData.Create( aCobDecoder : TCobDecoder;
                              aStream : IStream;
                              aPieceNames, aProcNames : TStringArray;
                              const aProcName : String;
                              aGlobalVarsCount : Integer);
begin
fCobDecoder := aCobDecoder;
PieceNames := aPieceNames;
ProcNames := aProcNames;
fProcName := aProcName;
fStream := aStream;
ExprStack := TStringlist.create;
ExprStack.sorted := false;
ExprStack.Duplicates := dupAccept;
DecompileAddresses := TList.create;
AddressesToVist := TStack.create;
Lines := TStringlist.create;
Lines.sorted := false;
Lines.Duplicates := dupAccept;
Params := TStringlist.create;
Params.sorted := false;
Params.Duplicates := dupError;
Globals := TStringlist.create;
Globals.sorted := false;
Globals.Duplicates := dupError;
while aGlobalVarsCount > 0 do
  begin
  dec( aGlobalVarsCount );
  Globals.Add( Format( 'Global_Var_%d', [ Globals.count+1 ] ));
  end;
end; {Create}

Destructor TProcData.Destroy;
begin
FreeAndNil( Globals );
FreeAndNil( Params );
FreeAndNil( Lines );
FreeAndNil( DecompileAddresses );
FreeAndNil( AddressesToVist );
FreeAndNil( ExprStack );
end; {Destroy}

Procedure TProcData.AddAddressToDecompileFrom( Offset : Longint );
var
  index : integer;
begin
assert( DecompileAddresses <> nil );
for index := 0 to DecompileAddresses.count-1 do
  if Offset = LongInt( DecompileAddresses[ index ] ) then
    exit; // duplicate address
DecompileAddresses.add( Pointer( Offset ) );
AddressesToVist.Push( Pointer( Offset ) );
end; {AddAddressToDecompileFrom}

Function TProcData.GetNextAddressToDecompileFrom : LongInt;
begin
assert( DecompileAddresses <> nil );
assert( AddressesToVist.count <> 0 );
Result := LongInt( AddressesToVist.Pop );
end; {GetNextAddressToDecompileFrom}

Function TProcData.IsMoreAddressToDecompile : boolean;
begin
result := AddressesToVist.count <> 0;
end; {IsMoreAddressToDecompile}

Function TProcData.IsInteger( const s : string; var Value : Integer ) : boolean;
var
  StartIndex : integer;
  Index : integer;
begin
if s = '' then
  begin
  result := false;
  exit;
  end;
if s[ 1 ] = '-' then
  StartIndex := 2
else
  StartIndex := 1;
for index := StartIndex to length(s) do
  if not (s[ index ] in [ '1'..'9','0' ]) then
    begin
    result := false;
    exit;
    end;
Value := StrToInt( s );
result := true;
end; {IsInteger}

Procedure TProcData.Decompile( StartOffset : Longint );
var
  index : integer;
  s : string;
begin
Lines.add( 'Begin' );
try
  AddAddressToDecompileFrom( StartOffset );
  while IsMoreAddressToDecompile do
    DecodeFromAddress( GetNextAddressToDecompileFrom );
finally
  if ExprStack.count <> 0 then
    begin
    Lines.Add('{');
    Lines.Add('*******************************************************');
    Lines.Add('* This script method has been decompiled incorrectly! *');
    Lines.Add('* Error: Data left on expression stack.               *');
    Lines.Add('*  This could be the result of a corrupted script Or  *');
    Lines.Add('*  bad decompilation.                                 *');
    Lines.Add('*******************************************************');
    Lines.Add('}');
    ExprStackDump;
    end;
  // biuld the param list
  if Params.count <> 0 then
    begin
    s := '( '+Params[ 0 ];
    for index := 1 to Params.count -1 do
      s := s +', ' +Params[ index ];
    s := s+' )';      
    end
  else  s := '()';
  Lines.Insert( 0, Format( 'function %s%s : LongInt;', [ ProcName, s ] ) );
  index := lines.count-1;
  // delete the last exit before the proc finishes
  if lines[ index ] = 'Exit;' then
    Lines.Delete( index );
  Lines.add( Format( 'End; {%s}', [ ProcName ] ) );
end;
end; {Decompile}

Procedure TProcData.DecodeFromAddress( StartOffset : Longint );
var
  value : longword;
  IsDone : boolean;
begin
// decompiles from an address
stream.Position := StartOffset;
repeat
  stream.ReadBuffer( value, sizeof( value ) );
  IsDone := DecodeOpcode( value );
  if ( AddressesToVist.count <> 0 ) And
     ( stream.Position >= LongInt( AddressesToVist.Examine ) ) then
    begin
    OutputLine( 'End;', [ ]{, false });
    IsDone := true;
    end;
until IsDone Or ( stream.position >= stream.size );
end; {DecodeFromAddress}

procedure TProcData.OutputLine( const s : string;
                                const args : array of const );
begin
Lines.AddObject( StringOfChar( ' ', 2*AddressesToVist.count )+Format( s, args ),
                 Pointer(Stream.position) );
end; {OutputLine}

Procedure TProcData.ExprStackDump;
var
  Index,count : integer;
begin
assert( self <> nil );
assert( ExprStack <> nil );
count := 0;
if ExprStack.Count = 0 then
  Lines.add( '// Expression stack empty.')
else
  For Index := ExprStack.Count-1 downto 0 do
    begin
    inc(count);
    Lines.add( Format( '// %d Expr: "%s", Priority:%d',
                       [ count, ExprStack[ Index ],
                         LongInt( ExprStack.Objects[ index ] ) ] ) );
    end;
end; {ExprStackDump}

Procedure TProcData.PushExpr( const Expr : string; ExprePriority : LongInt );
begin
assert( self <> nil );
assert( ExprStack <> nil );
ExprStack.AddObject( Expr, TObject( ExprePriority ) );
end; {PushExpr}

function TProcData.PopExpr( var Expr : string; ExprePriority : LongInt ) : LongInt;
var
  index : integer;
begin
assert( self <> nil );
assert( ExprStack <> nil );
if ExprStack.Count > 0 then
  begin
  index := ExprStack.Count-1;
  Expr := ExprStack[ index ];
  Result := LongInt( ExprStack.Objects[ index ] );
  if ExprePriority > LongInt( ExprStack.Objects[ index ] ) then
    Expr := '( '+Expr+' )';
  ExprStack.Delete( index );
  end
else
  Raise Exception.CreateFmt( ExprStackUnderFlow,
                             [ CobDecoder.ScriptName,ProcName ] );
end; {PopExpr}

Function TProcData.DecodeOpcode( aOpCode : Longword ) : boolean;

  Procedure ConvertLinear(var s : string);
  var
    value : LongInt;
  begin
  if IsInteger(s,value) then
    s := Format('[%.6f]',[value / Linear_Constant]);
  end; {ConvertToNorm}

  Procedure ConvertAngualar(var s : string);
  var
    value : LongInt;
  begin
  if IsInteger(s,value) then
    s := Format('<%.6f>',[value / Angular_Constant]);
  end; {ConvertAngualar}

const
  Axis : array [ 0..2 ] of string = ( 'x', 'y', 'z' );
var
  Priority,Arg1, Arg2 : LongInt;
  StackArg1, StackArg2,StackArg3,StackArg4,StackArg5 : string;
  IsInt1, IsInt2 : boolean;
  PrevPos,Index : LongInt;
  Res : string;
begin {DecodeOpcode}
Result := false;
Arg1 := 0;
Arg2 := 0;
case aOpCode of
  Opcode_MovePieceWithSpeed :
    begin
    PopExpr( StackArg2, ResOp );
    PopExpr( StackArg1, ResOp );
    stream.Read( Arg1, sizeof( Arg1 ) );
    stream.Read( Arg2, sizeof( Arg2 ) );
    ConvertLinear(StackArg1); ConvertLinear(StackArg2);
    OutputLine( 'MovePieceWithSpeed( %s, %s, %s, %s );',
                [ PieceNames[ Arg1 ], Axis[ Arg2 ], StackArg2, StackArg1 ] );
    end;
  Opcode_MovePieceNow :
    begin
    PopExpr( StackArg1, ResOp );
    stream.Read( Arg1, sizeof( Arg1 ) );
    stream.Read( Arg2, sizeof( Arg2 ) );
    ConvertLinear(StackArg1);
    OutputLine( 'MovePieceNow( %s, %s, %s );',
                [ PieceNames[ Arg1 ], Axis[ Arg2 ], StackArg1 ] );
    end;
  Opcode_WaitForMove :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    stream.Read( Arg2, sizeof( Arg2 ) );
    ConvertLinear(StackArg1); ConvertLinear(StackArg2);
    OutputLine( 'WaitForTurn( %s, %s );',
                [ PieceNames[ Arg1 ], Axis[ Arg2 ] ] );
    end;

  Opcode_TurnPieceWithSpeed :
    begin
    PopExpr( StackArg2, ResOp );
    PopExpr( StackArg1, ResOp );
    stream.Read( Arg1, sizeof( Arg1 ) );
    stream.Read( Arg2, sizeof( Arg2 ) );
    ConvertAngualar(StackArg1); ConvertAngualar(StackArg2);
    OutputLine( 'TurnPieceWithSpeed( %s, %s, %s, %s );',
                [ PieceNames[ Arg1 ], Axis[ Arg2 ], StackArg2, StackArg1 ] );
    end;
  Opcode_TurnPieceNow :
    begin
    PopExpr( StackArg1, ResOp );
    stream.Read( Arg1, sizeof( Arg1 ) );
    stream.Read( Arg2, sizeof( Arg2 ) );
    ConvertAngualar(StackArg1); ConvertAngualar(StackArg2);
    OutputLine( 'TurnPieceNow( %s, %s, %s );',
                [ PieceNames[ Arg1 ], Axis[ Arg2 ], StackArg1 ] );
    end;
  Opcode_StartSpin :
    begin
    PopExpr( StackArg2, ResOp );
    PopExpr( StackArg1, ResOp );
    stream.Read( Arg1, sizeof( Arg1 ) );
    stream.Read( Arg2, sizeof( Arg2 ) );
    ConvertAngualar(StackArg1); ConvertAngualar(StackArg2);
    OutputLine( 'StartSpin( %s, %s, %s, %s );',
                [ PieceNames[ Arg1 ], Axis[ Arg2 ], StackArg2, StackArg1 ] );
    end;
  Opcode_StopSpin :
    begin
    PopExpr( StackArg1, ResOp );
    stream.Read( Arg1, sizeof( Arg1 ) );
    stream.Read( Arg2, sizeof( Arg2 ) );
    ConvertAngualar(StackArg1); ConvertAngualar(StackArg2);    
    OutputLine( 'StopSpin( %s, %s, %s );',
                [ PieceNames[ Arg1 ], Axis[ Arg2 ], StackArg1 ] );
    end;
  Opcode_WaitForTurn :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    stream.Read( Arg2, sizeof( Arg2 ) );
    OutputLine( 'WaitForTurn( %s, %s );', [ PieceNames[ Arg1 ], Axis[ Arg2 ] ] );
    end;
                
  Opcode_ShowPiece :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( 'ShowPiece( %s );', [ PieceNames[ Arg1 ] ] );
    end;
  Opcode_HidePiece :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( 'HidePiece( %s );', [ PieceNames[ Arg1 ] ] );
    end;
  Opcode_CachePiece :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( 'CachePiece( %s );', [ PieceNames[ Arg1 ] ] );
    end;
  Opcode_DontCachePiece :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( 'DontCachePiece( %s );', [ PieceNames[ Arg1 ] ] );
    end;
  Opcode_DontShadow :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( 'DontShadow( %s );', [ PieceNames[ Arg1 ] ] );
    end;
  Opcode_DontShade :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( 'DontShade( %s );', [ PieceNames[ Arg1 ] ] );
    end;

  Opcode_Random :
    begin
    PopExpr( StackArg2, ResOp );
    PopExpr( StackArg1, ResOp );
    PushExpr( Format( 'Random( %s, %s )', [ StackArg1, StackArg2 ] ), ValueOp );
    end;
  Opcode_EmitSfx :
    begin
    PopExpr( StackArg1, ResOp );
    if IsInteger(StackArg1,Arg2) then
      begin
      if SFXTYPE_VTOL = Arg2  then
        StackArg1 := 'SFXTYPE_VTOL'
      else if SFXTYPE_THRUST = Arg2 then
        StackArg1 := 'SFXTYPE_THRUST'
      else if SFXTYPE_WAKE1 = Arg2  then
        StackArg1 := 'SFXTYPE_WAKE1'
      else if SFXTYPE_WAKE2 = Arg2  then
        StackArg1 := 'SFXTYPE_WAKE2'
      else if SFXTYPE_REVERSEWAKE1 = Arg2  then
        StackArg1 := 'SFXTYPE_REVERSEWAKE1'
      else if SFXTYPE_REVERSEWAKE2 = Arg2  then
        StackArg1 := 'SFXTYPE_REVERSEWAKE2'
      else if SFXTYPE_WHITESMOKE = Arg2 then
        StackArg1 := 'SFXTYPE_WHITESMOKE'
      else if SFXTYPE_BLACKSMOKE = Arg2 then
        StackArg1 := 'SFXTYPE_BLACKSMOKE'
      else if SFXTYPE_SUBBUBBLES = Arg2 then
        StackArg1 :=  'SFXTYPE_SUBBUBBLES'
      else Raise Exception.createFmt(UnknownSFXType,[Arg2]);
      end;
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( 'EmitSfx( %s, %s );', [ StackArg1, PieceNames[ Arg1 ] ] );
    end;
  Opcode_Explode :
    begin
    PopExpr( StackArg1, ResOp );
    if IsInteger(StackArg1,Arg2) then
      begin
      Arg2 := StrToInt(StackArg1);
      StackArg1 := '';
      if (SHATTER and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'SHATTER' ]);
      if (EXPLODE_ON_HIT and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'EXPLODE_ON_HIT' ]);
      if (FALL and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'FALL' ]);
      if (FIRE and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'FIRE' ]);
      if (BITMAPONLY and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'BITMAPONLY' ]);
      if (BITMAP1 and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'BITMAP1' ]);
      if (BITMAP2 and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'BITMAP2' ]);
      if (BITMAP3 and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'BITMAP3' ]);
      if (BITMAP4 and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'BITMAP4' ]);
      if (BITMAP5 and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'BITMAP5' ]);
      if (BITMAPNUKE and Arg2) <> 0 then
        StackArg1 := Format('%s Or %s', [ StackArg1, 'BITMAPNUKE' ]);
      Delete(StackArg1,1,4);
      end;
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( 'Explode( %s, %s );', [ PieceNames[ Arg1 ], StackArg1 ] );
    end;

  Opcode_Signal :
    begin
    PopExpr( StackArg1, ResOp );
    OutputLine( 'Signal( %s );', [ StackArg1 ] );
    end;
  Opcode_SetSignalMask :
    begin
    PopExpr( StackArg1, ResOp );
    OutputLine( 'SetSignalMask( %s );', [ StackArg1 ] );
    end;    
  Opcode_Sleep :
    begin // Set the time for when we are allowed to continue
    PopExpr( StackArg1, ResOp );
    OutputLine( 'Sleep( %s );', [ StackArg1 ] );
    end;


  // Stack ops
  Opcode_StackAllocate :
    begin // todo : implement code to determine how many f*cking parms & locals there are
    arg2 := 1;
    PrevPos := stream.Position;
    try
      repeat
        stream.Read( Arg1, sizeof( Arg1 ) );
        inc( arg2 );
      until Arg1 <> Opcode_StackAllocate;
      dec( arg2 );
      stream.Position := stream.Position - sizeof( Arg1 );
    except
      On EReadError do
        begin
        stream.Position := PrevPos;
        arg2 := 1;
        end;
    end;
//    OutputLine( 'StackAllocate( %d );', [ arg2 ] );
    while arg2 > 0 do
      begin
      dec( arg2 );
      Params.add( Format( 'Func_Var_%d', [ Params.count+1 ] ) );
      end;
    end;
  Opcode_PushConstantOntoStack :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    PushExpr( IntToStr( Integer( Arg1 ) ), ValueOp );
    end;
  Opcode_PushLocalVariableOntoStack :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    PushExpr( Params[ Arg1 ], ValueOp );
    end;
  Opcode_PushStaticVariableOntoStack :
    begin
    stream.Read( Arg1, sizeof( Arg1 ) );
    PushExpr( Globals[ Arg1 ], ValueOp );
    end;
  Opcode_SetLocalVariable :
    begin
    PopExpr( StackArg1, ResOp );
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( '%s := %s;', [ Params[ Arg1 ], StackArg1 ] );
    end;
  Opcode_SetStaticVariable :
    begin
    PopExpr( StackArg1, ResOp );
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( '%s := %s;', [ Globals[ Arg1 ], StackArg1 ] );
    end;
  Opcode_PopStack :
    begin
    asm int 3 end;
//    ExprStackDump;
    PopExpr( StackArg1, ValueOp );
    OutputLine( 'PopExpr;', [ ] );
    end;

  Opcode_SetUnitValue :
    begin
    Priority := PopExpr( StackArg2, ResOp );
    PopExpr( StackArg1, ResOp );
    Arg1 := StrToInt( StackArg1 );
    if not (SetAble in UnitProperties[ Arg1 ].Flags) then
      Raise Exception.CreateFmt( UnitPropertyNotSetable,[ UnitPropPrefix, UnitProperties[ Arg1 ].Name ]);
    assert(UnitProperties[ Arg1 ].argCount = 0);
    if (IsBool in UnitProperties[ Arg1 ].Flags) then
      begin
      if IsInteger( StackArg2, arg2 ) then
        begin
        if arg2 <> 0 then
          StackArg2 := 'True'
        else
          StackArg2 := 'False';
        end
      else if Priority > RelOp then
        StackArg2 := '( '+StackArg2 +' ) <> 0';
      end;
    OutputLine( '%s%s := %s;', [ UnitPropPrefix, UnitProperties[ Arg1 ].Name, StackArg2 ] );
    end;
  Opcode_GetUnitValue :
    begin
    PopExpr( StackArg1, ResOp );
    Arg1 := StrToInt( StackArg1 );
    assert(UnitProperties[ Arg1 ].argCount = 0);
    PushExpr( format( '%s%s', [ UnitPropPrefix, UnitProperties[ Arg1 ].Name ] ), ValueOp );
    end;
  Opcode_GetFunctionResult :
    begin
    // get the arguments of the expr stack
    // max of 4 for the property, 1 for which property
    PopExpr( StackArg5, ResOp );
    PopExpr( StackArg4, ResOp );
    PopExpr( StackArg3, ResOp );
    PopExpr( StackArg2, ResOp );
    PopExpr( StackArg1, ResOp );
    // determine which property we are using
    Arg1 := StrToInt( StackArg1 );
    Res := '';
    // trasform the individule arguments into a comma delimited argument list
    if UnitProperties[ Arg1 ].ArgCount >= 1 then
      begin
      if ArgIsPiece in UnitProperties[ Arg1 ].Flags then
        Res := Res + PieceNames[ StrToInt(StackArg2) ] + ', '
      else
        Res := Res + StackArg2 + ', ';
      end;
    if UnitProperties[ Arg1 ].ArgCount >= 2 then
      Res := Res + StackArg3 + ', ';
    if UnitProperties[ Arg1 ].ArgCount >= 3 then
      Res := Res + StackArg4 + ', ';
    if UnitProperties[ Arg1 ].ArgCount >= 4 then
      Res := Res + StackArg5 + ', ';
    // now actually generate the resulting source code  
    if length(Res) <> 0 then
      begin
      Setlength(Res,length(Res)-2);
      PushExpr( format( '%s%s( %s )',
                         [ UnitPropPrefix, UnitProperties[ Arg1 ].Name,Res ] ), ValueOp );
      end
    else
      PushExpr( format( '%s%s()',
                         [ UnitPropPrefix, UnitProperties[ Arg1 ].Name ] ), ValueOp );
    end;

  Opcode_StartScript : // starts the Proc as a new thread
    begin
    stream.Read( Arg1, sizeof( Arg1 ) ); stream.Read( Arg2, sizeof( Arg2 ) );
    // read params off the stack
    StackArg2 := '';
    while Arg2 > 0 do
      begin
      PopExpr( StackArg1, ResOp );
      StackArg2 := StackArg1+', '+StackArg2;
      dec( Arg2 );
      end;
    if length( StackArg2 ) <> 0 then
      begin
      setlength( StackArg2, length( StackArg2 )-2 );
      OutputLine( 'StartScript( %s, [ %s ] );', [ ProcNames[ Arg1 ], StackArg2 ] );
      end
    else
      OutputLine( 'StartScript( %s );', [ ProcNames[ Arg1 ] ] );
    end;
  Opcode_CallScript :  // Normal Proc call
    begin
    stream.Read( Arg1, sizeof( Arg1 ) ); stream.Read( Arg2, sizeof( Arg2 ) );
    // read params off the stack ( from right to left )
    StackArg2 := '';
    while Arg2 > 0 do
      begin
      PopExpr( StackArg1, ResOp );
      StackArg2 := StackArg1+', '+StackArg2;
      dec( Arg2 );
      end;
    if length( StackArg2 ) <> 0 then
      begin
      setlength( StackArg2, length( StackArg2 )-2 );
      OutputLine( '%s( %s );', [ ProcNames[ Arg1 ], StackArg2 ] );
      end
    else
      OutputLine( '%s();', [ ProcNames[ Arg1 ] ] );
    end;

  Opcode_Return :
    begin
    PopExpr( StackArg1, ResOp );
    OutputLine( 'Result := %s;', [ StackArg1 ] );
    OutputLine( 'Exit;', [ ]{, false} );
    // signal that the decoding is done.
    Result := true;
    end;
  Opcode_JumpToOffset :
    begin
//    Thread.Offset = pCode[ Thread.Offset + 1 ];
    stream.Read( Arg1, sizeof( Arg1 ) );
    Arg1 := Arg1 * 4;
    if Arg1 < Stream.Position - sizeof(longword) * 2 then
      begin // do fixup on existing source code
      for index := 0 to Lines.Count-1 do
        begin
        if LongInt(Lines.Objects[ index ]) > Arg1 then
          begin
          StackArg1 := Lines[ index ];
          If AnsiPos('if',StackArg1) <> 0 then
            begin
            // change the start marker
            StackArg1 := StringReplace(StackArg1,'if','while',[ rfIgnoreCase ]);
            // change the end marker
            StackArg1 := StringReplace(StackArg1,'then','do',[ rfIgnoreCase ]);
            // repack the modified line
            Lines[ index ] := StackArg1;
            end
          else
            raise Exception.create('Could not find start of while loop marker');
          break;
          end;
        end;
      end
    else
      begin
      ExprStackDump;
      raise Exception.create('Unknown loop style');
      end;
    end;
  Opcode_JumpToOffsetIfFalse :
    begin
    Priority := PopExpr( StackArg1, ResOp );
    stream.Read( Arg1, sizeof( Arg1 ) );
    Arg1 := Arg1 * 4;
    if (StackArg1[ 1 ] <> '(') and
       (IsInteger( StackArg1, Arg2 )) then
      begin
      if Arg2 <> 0 then
        OutputLine( 'if True then', [ ] )
      else
        OutputLine( 'if False then', [ ] );
      end
    else
      begin
      if Priority = RelOp then
        OutputLine( 'if %s then', [ StackArg1 ] )
      else
        OutputLine( 'if %s <> 0 then', [ StackArg1 ] );
      end;
    // tag the dest address to be decompiled
    AddAddressToDecompileFrom( Arg1 );
    // output filler source code line
    OutputLine( 'Begin', [ ] );
    end;

  Opcode_AttachUnit :
    begin
    // not used! Always Zero.
    PopExpr( StackArg3, ResOp );
    // The piece to use
    PopExpr( StackArg2, ResOp );
    // the unit ID
    PopExpr( StackArg1, ResOp );
    Arg1 := StrToInt( StackArg2 );
    if (Arg1 >= low(PieceNames)) and (Arg1 <= high(PieceNames)) then
      OutputLine( 'AttachUnit( %s, %s );', [ StackArg1, PieceNames[ Arg1 ] ] )
    else if Arg1 = -1 then
      OutputLine( 'AttachUnit( %s, '''' );', [ StackArg1 ] )
    else
      Raise Exception.create(InvaliePieceIndex);
    end;
  Opcode_DropUnit :
    begin
    // the unit ID
    PopExpr( StackArg1, ResOp );
    OutputLine( 'DropUnit( %s );', [ StackArg1 ] );
    end;
        
  Opcode_Unknown1, 
  Opcode_Unknown2,
  Opcode_Unknown3 :
    begin
    ExprStackDump;
    asm int 3 end;    
    Result := true;
    OutputLine('Unknown opcode: $%s',[ InttoHex( aOpCode, 8 ) ]);
    end;
  Opcode_MapCommand :
    begin
    ExprStackDump;
    asm int 3 end;
    stream.Read( Arg1, sizeof( Arg1 ) );
    stream.Read( Arg1, sizeof( Arg1 ) );
    OutputLine( 'MapCommand( %d, %d );', [ Arg1, Arg2 ] );
    end;


  Opcode_Add :
    begin // Operation( + );
    PopExpr( StackArg2, AddOp );
    PopExpr( StackArg1, AddOp );
    IsInt1 := IsInteger( StackArg1, arg1 );
    IsInt2 := IsInteger( StackArg2, arg2 );
    if IsInt1 And IsInt2 then
      PushExpr( IntToStr( arg1 + arg2 ), ValueOp )
    else if IsInt1 and ( arg1 = 0 ) then
      PushExpr( StackArg2, ValueOp )
    else if IsInt2 and ( arg2 = 0 ) then
      PushExpr( StackArg1, ValueOp )
    else
      PushExpr( Format( '%s + %s', [ StackArg1, StackArg2 ] ), AddOp )
    end;
  Opcode_Subtract :
    begin // Operation( - );
    PopExpr( StackArg2, AddOp );
    PopExpr( StackArg1, AddOp );
    IsInt1 := IsInteger( StackArg1, arg1 );
    IsInt2 := IsInteger( StackArg2, arg2 );
    if IsInt1 And IsInt2 then
      PushExpr( IntToStr( arg1 - arg2 ), ValueOP )
    else if IsInt1 and ( arg1 = 0 ) then
      PushExpr( '- '+StackArg2, ValueOP )
    else if IsInt2 and ( arg2 = 0 ) then
      PushExpr( StackArg1, ValueOP )
    else
      PushExpr( Format( '%s - %s', [ StackArg1, StackArg2 ] ), AddOp )
    end;
  Opcode_Multiply :
    begin // Operation( * );
    PopExpr( StackArg2, MulOp );    
    PopExpr( StackArg1, MulOp );
    IsInt1 := IsInteger( StackArg1, arg1 );
    IsInt2 := IsInteger( StackArg2, arg2 );
    if IsInt1 And IsInt2 then
      PushExpr( IntToStr( arg1 * arg2 ), ValueOp )
    else if ( IsInt1 and ( arg1 = 0 ) ) Or
            ( IsInt2 and ( arg2 = 0 ) ) then
      PushExpr( '0', ValueOp )
    else
      PushExpr( Format( '%s * %s', [ StackArg1, StackArg2 ] ), MulOp )
    end;
  Opcode_Divide :
    begin // Operation( / );
    PopExpr( StackArg2, MulOp );    
    PopExpr( StackArg1, MulOp );
    IsInt1 := IsInteger( StackArg1, arg1 );
    IsInt2 := IsInteger( StackArg2, arg2 );
    if IsInt1 And IsInt2 then
      PushExpr( IntToStr( arg1 div arg2 ), ValueOp )
    else if IsInt1 and ( arg1 = 0 ) then
      PushExpr( '0', ValueOp )
    else if IsInt2 and ( arg2 = 0 ) then
      PushExpr( StackArg1+' div 0{Error: Divide by zero}',ValueOp )
    else
      PushExpr( Format( '%s / %s', [ StackArg1, StackArg2 ] ), MulOp )
    end;
  Opcode_BitwiseAnd :
    begin // Operation( & );
    PopExpr( StackArg2, MulOp );    
    PopExpr( StackArg1, MulOp );
    IsInt1 := IsInteger( StackArg1, arg1 );
    IsInt2 := IsInteger( StackArg2, arg2 );
    if IsInt1 And IsInt2 then
      PushExpr( IntToStr( arg1 and arg2 ), ValueOp )
    else if ( IsInt1 and ( arg1 = 0 ) ) Or
            ( IsInt2 and ( arg2 = 0 ) ) then
      PushExpr( '0', ValueOp )
    else
      PushExpr( Format( '%s And %s', [ StackArg1, StackArg2 ] ), MulOp )
    end;
  Opcode_BitwiseOr :
    begin  // Operation( | );
    PopExpr( StackArg2, AddOp );    
    PopExpr( StackArg1, AddOp );
    IsInt1 := IsInteger( StackArg1, arg1 );
    IsInt2 := IsInteger( StackArg2, arg2 );
    if IsInt1 And IsInt2 then
      PushExpr( IntToStr( arg1 Or arg2 ), ValueOp )
    else if IsInt1 and ( arg1 = 0 ) then
      PushExpr( StackArg2, ValueOp )
    else if IsInt2 and ( arg2 = 0 ) then
      PushExpr( StackArg1, ValueOp )
    else
      PushExpr( Format( '%s Or %s', [ StackArg1, StackArg2 ] ), AddOp )
    end;
  Opcode_LessThan :
    begin // Operation( < );
    PopExpr( StackArg2, RelOp );
    PopExpr( StackArg1, RelOp );
    if IsInteger( StackArg1, arg1 ) And IsInteger( StackArg2, arg2 ) then
      PushExpr( BoolToStr( arg1 < arg2 ), ValueOp )
    else
      PushExpr( Format( '%s < %s', [ StackArg1, StackArg2 ] ), RelOp );
    end;
  Opcode_LessThanOrEqual :
    begin // Operation( <= );
    PopExpr( StackArg2, RelOp );
    PopExpr( StackArg1, RelOp );
    if IsInteger( StackArg1, arg1 ) And IsInteger( StackArg2, arg2 ) then
      PushExpr( BoolToStr( arg1 <= arg2 ), ValueOp )
    else
      PushExpr( Format( '%s <= %s', [ StackArg1, StackArg2 ] ), RelOp );
    end;
  Opcode_GreaterThan :
    begin // Operation( > );
    PopExpr( StackArg2, RelOp );
    PopExpr( StackArg1, RelOp );
    if IsInteger( StackArg1, arg1 ) And IsInteger( StackArg2, arg2 ) then
      PushExpr( BoolToStr( arg1 > arg2 ), ValueOp )
    else
      PushExpr( Format( '%s > %s', [ StackArg1, StackArg2 ] ), RelOp );
    end;
  Opcode_GreaterThanOrEqual :
    begin // Operation( >= );
    PopExpr( StackArg2, RelOp );
    PopExpr( StackArg1, RelOp );
    if IsInteger( StackArg1, arg1 ) And IsInteger( StackArg2, arg2 ) then
      PushExpr( BoolToStr( arg1 >= arg2 ), ValueOp )
    else
      PushExpr( Format( '%s >= %s', [ StackArg1, StackArg2 ] ), RelOp );
    end;
  Opcode_Equal :
    begin // Operation( = );
    PopExpr( StackArg2, RelOp );
    PopExpr( StackArg1, RelOp );
    if IsInteger( StackArg1, arg1 ) And IsInteger( StackArg2, arg2 ) then
      PushExpr( BoolToStr( arg1 = arg2 ), ValueOp )
    else
      PushExpr( Format( '%s = %s', [ StackArg1, StackArg2 ] ), RelOp );
    end;
  Opcode_NotEqual :
    begin // Operation( != );
    PopExpr( StackArg2, RelOp );
    PopExpr( StackArg1, RelOp );
    if IsInteger( StackArg1, arg1 ) And IsInteger( StackArg2, arg2 ) then
      PushExpr( BoolToStr( arg1 <> arg2 ), ValueOp )
    else  
      PushExpr( Format( '%s <> %s', [ StackArg1, StackArg2 ] ), RelOp );
    end;
  Opcode_And :
    begin // Operation( && );
    PopExpr( StackArg2, MulOp );
    PopExpr( StackArg1, MulOp );
    IsInt1 := IsInteger( StackArg1, arg1 );
    IsInt2 := IsInteger( StackArg2, arg2 );
    if IsInt1 And IsInt2 then
      PushExpr( BoolToStr( ( arg1 And arg2 ) <> 0 ), ValueOp )
    else if ( IsInt1 and ( arg1 = 0 ) ) Or
            ( IsInt2 and ( arg2 = 0 ) ) then
      PushExpr( 'False', ValueOp )
    else
      PushExpr( Format( '%s And %s', [ StackArg1, StackArg2 ] ), MulOp );
    end;
  Opcode_Or :
    begin // Operation( || );
    PopExpr( StackArg2, AddOp );    
    PopExpr( StackArg1, AddOp );
    IsInt1 := IsInteger( StackArg1, arg1 );
    IsInt2 := IsInteger( StackArg2, arg2 );
    if IsInt1 And IsInt2 then
      PushExpr( BoolToStr( ( arg1 Or arg2 ) <> 0 ),ValueOp )
    else if IsInt1 and ( arg1 = 0 ) then
      PushExpr( StackArg2, ValueOp )
    else if IsInt2 and ( arg2 = 0 ) then
      PushExpr( StackArg1, ValueOp )
    else
      PushExpr( Format( '%s Or %s', [ StackArg1, StackArg2 ] ), AddOp );
    end;
  Opcode_Not :
    begin // Operation( not );
    PopExpr( StackArg1, NotOp );
    if IsInteger( StackArg1, arg1 ) then
      PushExpr( IntTostr( Not arg1 ), ValueOp )
    else
      PushExpr( Format( 'Not %s', [ StackArg1 ] ), NotOp );
    end;
else // error unknown opcode!
  asm int 3 end;    
end; {Case}
end; {DecodeOpcode}

// -----------------------------------------------------------------------------
//  TCobDecoder
// -----------------------------------------------------------------------------

Constructor TCobDecoder.Create( aStream : IStream; const aScriptName : string );
begin
Inherited create;
fScriptName := aScriptName;
Stream := aStream;
// create the output stream
OutputStream := TStringStream.create( '' );
// read the cob header
ReadHeader;
// decode the cob procs
DecodeProcs;
end; {Create}

Destructor TCobDecoder.Destroy;
begin
OutputStream := nil;
Inherited;
end; {Destroy}

function TCobDecoder.ReadString( Offset : Longint ) : string;
var
  OldOffset : integer;
  Buffer : Array of byte;
begin
if ( Offset >= Stream.size ) Or
   ( Stream.size = Stream.Position ) then
  begin
  result := '';
  exit;
  end;
OldOffset := Stream.Position;
try
  Stream.Position := Offset;
  Setlength( buffer, Stream.size - Stream.Position );
  Stream.ReadBuffer( buffer[ 0 ], length( buffer ) );
  Result := pchar( @buffer[ 0 ] );
  buffer := nil;
finally
  Stream.Position := OldOffset;
end;
end; {ReadString}

function TCobDecoder.ReadOffsetArray( Size, Offset : Longint ) : TOffsetArray;
var
  OldOffset : integer;
begin
if ( size<> 0 ) And
   ( Offset < Stream.size ) then
  begin
  OldOffset := Stream.Position;
  try
    Stream.Position := Offset;
    setlength( result, size );
    Stream.ReadBuffer( result[ 0 ], size * sizeof( result[ 0 ] ) );
  finally
    Stream.Position := OldOffset;
  end;
  end
else result := nil;
end; {ReadOffsetArray}

function TCobDecoder.ReadStringArray( Size, Offset : Longint ) : TStringArray;
var
  OffsetArray :TOffsetArray;
  index : integer;
begin
OffsetArray := nil;
if ( size<> 0 ) And
   ( Offset < Stream.size ) then
  begin
  setlength( result ,size);
  OffsetArray := ReadOffsetArray( Size, Offset );
  for index := low( OffsetArray ) to high( OffsetArray ) do
    result[ index ] := ReadString( OffsetArray[ index ] );
  end
else result := nil;
end; {ReadStringArray}

Procedure TCobDecoder.ReadHeader;
var
  CobHeader : TCobHeader;
begin
// read in the header
stream.ReadBuffer( CobHeader, sizeof( CobHeader ) );
case CobHeader.Version of
  CobVersion_TotalAnnihilation : ;
  CobVersion_Kingdoms : Raise EScriptingError.create( 'TA:K COB scripts are not supported' );
else Raise EScriptingError.create( 'Unknown COB verison' );
end; {case}
// load the various name arrays
PieceNames := ReadStringArray( CobHeader.NumberOfPieces, CobHeader.OffsetToPieceNameOffsetArray );
ProcNames := ReadStringArray( CobHeader.NumberOfProcs, CobHeader.OffsetToProcNameOffsetArray );
GlobalVarsCount := CobHeader.NumberOfStaticVars;
ReadProcs( CobHeader );
end; {ReadHeader}

Procedure TCobDecoder.ReadProcs( const CobHeader : TCobHeader );
var
  ProcOffsets : TOffsetArray;
  index : integer;
begin
ProcOffsets := ReadOffsetArray( CobHeader.NumberOfProcs, CobHeader.OffsetToProcPointerArray );
if ( ProcOffsets= nil ) Or ( length( ProcOffsets ) = 0 ) then exit;
setlength( ByteCode, CobHeader.LengthOfAllProcs );
if ( ByteCode = nil ) Or ( length( ByteCode ) = 0 ) then  exit;
if ( CobHeader.OffsetToFirstProc + length( ByteCode ) >= stream.size ) then exit;
stream.position := CobHeader.OffsetToFirstProc;
stream.ReadBuffer( ByteCode[ 0 ], length( ByteCode ) * sizeof( ByteCode[ 0 ] ) );
setlength( Procs, CobHeader.NumberOfProcs );
for index := 0 to high( Procs ) do
  begin
  Procs[ index ].Name := ProcNames[ index ];
  Procs[ index ].Offset := ProcOffsets[ index ] * sizeof( longword );
  end;
end; {ReadProcs}

Procedure TCobDecoder.DecodeProcs;
var
  index : integer;
  mem : IStream;
begin
mem := TMemoryStream.create;
Mem.WriteBuffer( ByteCode[ 0 ], length( ByteCode ) * sizeof( ByteCode[ 0 ] ) );
for index := 0 to high( Procs ) do
  DecodeProc( Mem, Procs[ index ] );
end; {DecodeProcs}

Procedure SaveOutput( stream : IStream; const ScriptName,ProcName : string);
var
  fs : IStream;
begin
stream.Position := 0;
fs := TFileStream.create( Format( 'scripts\%s_Proc_%s.txt',
                                  [ ScriptName,ProcName ]),
                          [ fmCreate ] );
fs.copyFrom( stream, 0 );
end; {SaveOutput}

Procedure TCobDecoder.DecodeProc( stream : IStream;const CobProc : TCobProc );
var
  ProcData : TProcData;
  Index : integer;
//  CodeRefactoriser : TCodeRefactoriser;  
begin {DecodeProc}
ProcData := TProcData.create( self, stream,
                              PieceNames, ProcNames,
                              CobProc.Name, GlobalVarsCount);
try
//  try
    ProcData.Decompile( CobProc.Offset );
//  except
    for Index := 0 to ProcData.Lines.count -1 do
      OutputStream.WriteString( ProcData.Lines[ Index ] + NewLine);
    SaveOutput( OutputStream, ScriptName, CobProc.Name );
    OutputStream.Size := 0;
{
    raise;
  end;
  CodeRefactoriser := TCodeRefactoriser.create( ProcData.Lines,
                                                ProcData.Globals,
                                                ProcData.Params,
                                                ProcData.Locals);
  for Index := 0 to CodeRefactoriser.Lines.count -1 do
    Output.WriteString(CodeRefactoriser.Lines[ Index ]+NewLine);
  SaveOutput(Output,ScriptName,CobProc.Name);
  Output.Size := 0;
}
finally
  ProcData.free;
end;
end; {DecodeProc}

end.
